package net.knarcraft.minecraftserverlauncher.profile;

import net.knarcraft.minecraftserverlauncher.server.ServerTypeHandler;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerLauncherGUI;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerConsoles;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerTab;
import net.knarcraft.minecraftserverlauncher.server.Server;
import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.utility.JarDownloader;

import javax.naming.ConfigurationException;
import javax.swing.*;
import java.io.*;
import java.util.ArrayList;
import java.util.NoSuchElementException;
import java.util.Scanner;
import java.util.concurrent.Executors;

/**
 * Contains all user settings, and a list of servers.
 *
 * @author      Kristian Knarvik <kristian.knarvik@knett.no>
 * @version     1.0.0
 * @since       1.0.0
 */
public class Profile {
    private static final ArrayList<Profile> profiles = new ArrayList<>();
    private static Profile current;
    private static ServerLauncherGUI serverLauncherGui;
    private static final String profilesDir = Main.getApplicationWorkDirectory() + File.separator +  "files";
    private static final String profilesFile = Main.getApplicationWorkDirectory() + File.separator +  "files" + File.separator + "Profiles.txt";
    private static final String jarDirectory = Main.getApplicationWorkDirectory() + File.separator + "files" + File.separator + "Jars" + File.separator;

    private final ArrayList<Collection> collections;
    private final String name;
    private boolean runInBackground;
    private int delayStartup;
    private boolean downloadAllAvailableJARFiles;
    private static String vanillaVersion;
    private static String snapshotVersion;
    private static String bungeeVersion;

    private Profile(String name) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = false;
        this.delayStartup = 0;
        this.downloadAllAvailableJARFiles = false;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    private Profile(String name, boolean runInBackground, int delayStartup, boolean downloadAllAvailableJARFiles) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = runInBackground;
        this.delayStartup = delayStartup;
        this.downloadAllAvailableJARFiles = downloadAllAvailableJARFiles;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    public static ServerLauncherGUI getGUI() {
        return serverLauncherGui;
    }

    public boolean getRunInBackground() {
        return this.runInBackground;
    }

    public int getDelayStartup() {
        return this.delayStartup;
    }

    public boolean getDownloadAllAvailableJARFiles() {
        return this.downloadAllAvailableJARFiles;
    }

    public static Profile getCurrent() {
        return current;
    }

    public ArrayList<Collection> getCollections() {
        return this.collections;
    }

    public String getName() {
        return this.name;
    }

    /**
     * Gets a Collection object by name.
     *
     * @param name  The name of the collection.
     * @return      A collection object.
     */
    public Collection getCollection(String name) {
        for (Collection collection : this.collections) {
            if (collection.getName().equals(name)) {
                return collection;
            }
        }
        return null;
    }

    public static Profile getProfile(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                return profile;
            }
        }
        return null;
    }

    public static ArrayList<Profile> getProfiles() {
        return profiles;
    }

    public void setRunInBackground(boolean value) {
        this.runInBackground = value;
    }

    public void setDelayStartup(int value) {
        if (value >= 0) {
            this.delayStartup = value;
        }
    }

    public void setDownloadAllAvailableJARFiles(boolean value) {
        this.downloadAllAvailableJARFiles = value;
    }

    /**
     * Set the current profile to the profile with a certain name.
     *
     * @param name  The name of the profile
     */
    public static  void setCurrent(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                current = profile;
                break;
            }
        }
    }

    /**
     * Adds a collection to the profile if the name is valid.
     *
     * @param name  The name of the collection and its elements.
     */
    public void addCollection(String name) throws ConfigurationException {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (getCollection(name) == null &&
                !name.equals("") &&
                !name.equals("All") &&
                !name.contains("!") &&
                !name.contains("?") &&
                !name.contains(";")
                ) {
            collections.add(new Collection(name));
        } else {
            serverLauncherGui.showError("A server name must my unique and not empty or \"All\"." +
                    "It can't contain any of the characters \"!\", \"?\" or \";\".");
        }
    }

    /**
     * Adds a profile if the name is valid and unique.
     *
     * @param name  The name of the new profile.
     */
    public static void addProfile(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (name.equals("") && !name.matches("^[!?;]+$")) {
            serverLauncherGui.showError("Profile name can't be blank.");
            return;
        }
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                serverLauncherGui.showError("There is already a profile with this name.");
                return;
            }
        }
        new Profile(name);
    }

    public void removeCollection(String name) {
        for (int i = 0; i < collections.size(); i++) {
            if (collections.get(i).getName().equals(name)) {
                this.collections.remove(i);
                serverLauncherGui.update();
                break;
            }
        }
    }

    /**
     * Removes a profile with the given name from the list of profiles, if such a profile exists
     * @param name <p>The name of the profile to rempve</p>
     */
    public static void removeProfile(String name) {
        if (profiles.size() > 1) {
            profiles.removeIf(profile -> profile.name.equals(name));
        }
    }

    public void updateConsoles() {
        JTabbedPane consolesTab = ServerConsoles.getTabbedPane();
        consolesTab.removeAll();
        for (Collection collection : collections) {
            consolesTab.add(collection.getName(), collection.getServerConsole().getPanel());
        }
    }

    /**
     * Sends a command to a server, or all servers
     *
     * @param serverName    The target server
     * @param command       The command to send.
     */
    public void sendCommand(String serverName, String command) {
        if (serverName.equals("All")) {
            for (Collection collection : this.collections) {
                try {
                    collection.getServer().sendCommand(command);
                } catch (IOException e) {
                    serverLauncherGui.showError("Server " + collection.getName() + " caused an exception.");
                }
            }
        } else {
            Collection collection = getCollection(serverName);
            if (collection != null) {
                Server target = collection.getServer();
                try {
                    target.sendCommand(command);
                } catch (IOException e) {
                    serverLauncherGui.showError("Server " + target.getName() + " caused an exception.");
                }
            } else {
                serverLauncherGui.showError("Server " + serverName + " is invalid.");
            }
        }
    }

    /**
     * Reads all server tabs, and saves it to the variables of the corresponding servers.
     * Saves all profiles and servers to a text file.
     */
    public void save() throws FileNotFoundException {
        for (Collection collection : this.collections) {
            Server server = collection.getServer();
            ServerTab serverTab = collection.getServerTab();
            server.setPath(serverTab.getPath());
            server.setMaxRam(serverTab.getMaxRam());
            server.setType(ServerTypeHandler.getByName(serverTab.getType()));
            try {
                server.setServerVersion(serverTab.getVersion());
            } catch (IllegalArgumentException e) {
                serverLauncherGui.showError("Invalid server version for " + server.getName());
            }
            server.toggle(serverTab.enabled());
        }
        if (!new File(profilesDir).exists() && !new File(profilesDir).mkdirs()) {
            serverLauncherGui.showError("Unable to create the folder " + profilesDir);
            throw new FileNotFoundException("Unable to create the profiles folder: " + profilesDir);
        }
        try (PrintWriter file = new PrintWriter(profilesFile)) {
            int width;
            int height;
            if (serverLauncherGui == null) {
                width = 440;
                height = 170;
            } else {
                width = serverLauncherGui.getSize().width;
                height = serverLauncherGui.getSize().height;
            }
            file.println(String.format(
                    "%s;%s;%s;%s;%d;%d",
                    current.name,
                    vanillaVersion,
                    snapshotVersion,
                    bungeeVersion,
                    width,
                    height
            ));
            file.close();
            for (Profile profile : profiles) {
                StringBuilder saveString = new StringBuilder(String.format(
                        "%s;%b;%d;%b?",
                        profile.name,
                        profile.runInBackground,
                        profile.delayStartup,
                        profile.downloadAllAvailableJARFiles)
                );
                for (Collection collection : profile.collections) {
                    Server server = collection.getServer();
                    saveString.append(String.format(
                            "%s;%s;%b;%s;%s;%s;%s;%s;%s;%s!",
                            server.getName(),
                            server.getPath(),
                            server.isEnabled(),
                            server.getTypeName(),
                            server.getServerVersion(),
                            server.getMaxRam(),
                            server.getVanillaVersion(),
                            server.getSnapshotVersion(),
                            server.getSpongeVanillaVersion(),
                            server.getBungeeVersion()
                            )
                    );
                }
                saveString = new StringBuilder(saveString.substring(0, saveString.length() - 1));
                try (PrintWriter fileAppend = new PrintWriter(new FileWriter(
                        profilesFile,
                        true
                ))) {
                    fileAppend.println(saveString);
                } catch (IOException e) {
                    if (serverLauncherGui != null) {
                        serverLauncherGui.showError("Unable to save to file. Try running the software as an administrator.");
                    } else {
                        System.out.println("Unable to save to file. Try running the software as an administrator.");
                    }
                    throw new FileNotFoundException("Unable to save to the profiles file.");
                }
            }
        } catch (IOException e) {
            if (serverLauncherGui != null) {
                serverLauncherGui.showError("Unable to save to file. Try running the software as an administrator.");
            }
            throw new FileNotFoundException("Unable to create the profiles file");
        }
    }

    /**
     * Reads profiles and servers from a text file.
     */
    public static void load() throws ConfigurationException {
        try (Scanner in = new Scanner(new File(profilesFile))) {
            try {
                String[] staticData = in.nextLine().split(";", -1);
                String profileName = staticData[0];
                vanillaVersion = staticData[1];
                snapshotVersion = staticData[2];
                bungeeVersion = staticData[3];
                int guiWidth = Integer.parseInt(staticData[4]);
                int guiHeight = Integer.parseInt(staticData[5]);
                serverLauncherGui = new ServerLauncherGUI(guiWidth, guiHeight);
                while (in.hasNextLine()) {
                    String line = in.nextLine();
                    if (line.contains("?")) {
                        String[] data = line.split("\\?");
                        String[] profileData = data[0].split(";", -1);
                        Profile profile = parseProfile(profileData);
                        if (data[1].contains("!")) {
                            String[] servers = data[1].split("!", -1);
                            for (String server : servers) {
                                String[] serverData = server.split(";", -1);
                                parseServer(profile, serverData);
                            }
                        } else {
                            String[] serverData = data[1].split(";", -1);
                            parseServer(profile, serverData);
                        }
                    } else {
                        String[] profileData = line.split(";", -1);
                        parseProfile(profileData);
                    }
                }
                current = getProfile(profileName);
            } catch (ArrayIndexOutOfBoundsException | NumberFormatException e) {
                e.printStackTrace();
                serverLauncherGui.showError("Invalid Profile.txt file. Profiles could not be loaded. If this error persists, please " +
                        "manually delete the file.");
                System.exit(1);
            } catch (IOException e) {
                e.printStackTrace();
            }
            if (profiles.size() == 0) {
                addProfile("Default");
            }
        } catch (FileNotFoundException | NoSuchElementException e) {
            serverLauncherGui.showMessage("A profiles file was not found. Default profile was created.");
            try {
                serverLauncherGui = new ServerLauncherGUI();
            } catch (FileNotFoundException ex) {
                serverLauncherGui.showMessage("Failed to load ServerLauncherGUI messages. The ServerLauncherGUI can't be shown.");
            } catch (IOException ex) {
                ex.printStackTrace();
            }
            addProfile("Default");
        }
        serverLauncherGui.update();
        serverLauncherGui.updateProfiles();
        current.updateConsoles();
        if (current.downloadAllAvailableJARFiles) {
            Executors.newSingleThreadExecutor().execute(() -> {
                JarDownloader downloader = new JarDownloader(serverLauncherGui, jarDirectory);
                try {
                    downloader.downloadJars();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            });
        }
        if (current.runInBackground) {
            Executors.newSingleThreadExecutor().execute(Server::startServers);
            serverLauncherGui.hide();
        }
    }

    /**
     * Parses a profile, and creates a profile with the data.
     * @param profileData <p>The data of the new profile</p>
     * @return <p>The new profile</p>
     */
    private static Profile parseProfile(String[] profileData) {
        return new Profile(
                profileData[0],
                Boolean.parseBoolean(profileData[1]),
                Integer.parseInt(profileData[2]),
                Boolean.parseBoolean(profileData[3])
        );
    }

    /**
     * Parses a server, and creates a new collection.
     * @param profile <p>The profile which to add the collection</p>
     * @param serverData <p>The data to parse</p>
     */
    private static void parseServer(Profile profile, String[] serverData) throws ConfigurationException {
        profile.collections.add(new Collection(
                serverData[0],
                serverData[1],
                Boolean.parseBoolean(serverData[2]),
                serverData[3],
                serverData[4],
                serverData[5],
                serverData[6],
                serverData[7],
                serverData[8],
                serverData[9])
        );
    }
}
