package net.knarcraft.minecraftserverlauncher.server;

import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.profile.Collection;
import net.knarcraft.minecraftserverlauncher.profile.Profile;
import net.knarcraft.minecraftserverlauncher.server.servertypes.ServerType;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.concurrent.TimeUnit;


/**
 * Contains all necessary information to create, run and manage a Minecraft server.
 *
 * @author Kristian Knarvik <kristian.knarvik@knett.no>
 * @version 1.0.0
 * @since 1.0.0
 */
public class Server {
    /**
     * Available ram sizes. For ServerLauncherGUI dropdown
     */
    private static final String[] ramList = {
            "512M", "1G", "2G", "3G", "4G", "5G", "6G", "7G", "8G", "9G", "10G", "11G", "12G", "13G", "14G", "15G", "16G"
    };
    private static final String jarDirectory = Main.getApplicationWorkDirectory() + File.separator + "files" + File.separator + "Jars" + File.separator;

    private final String name;
    private String path;
    private boolean enabled;
    private final ArrayList<String> playerList;
    private ServerType type;
    private String serverVersion;
    private String maxRam;
    private Process process;
    private BufferedWriter writer;
    private BufferedReader reader;
    private final String vanillaVersion;
    private final String snapshotVersion;
    private final String spongeVanillaVersion;
    private final String bungeeVersion;
    private boolean started;

    /**
     * Initializes a new server with default values
     *
     * @param name <p>The name of the server</p>
     */
    public Server(String name) {
        this.name = name;
        this.path = "";
        this.enabled = false;
        this.playerList = new ArrayList<>();
        this.type = null;
        this.serverVersion = "";
        this.maxRam = ramList[0];
        this.process = null;
        this.writer = null;
        this.reader = null;
        this.vanillaVersion = "";
        this.snapshotVersion = "";
        this.spongeVanillaVersion = "";
        this.bungeeVersion = "";
    }

    /**
     * Initializes a server with the given values
     *
     * @param name <p>The name of the server</p>
     * @param path <p>The file path of the folder containing the server files</p>
     * @param enabled <p>Whether the server is enabled to start the next time servers are started</p>
     * @param typeName <p>The name of the server type currently in use on the server</p>
     * @param serverVersion <p>The currently selected server version for the given server type</p>
     * @param maxRam <p>The maximum amount of ram the server is allowed to use</p>
     * @param vanillaVersion <p>The version of the "latest" downloaded vanilla version</p>
     * @param snapshotVersion <p>The version of the "latest" downloaded snapshot version</p>
     * @param spongeVanillaVersion <p>The version of the "latest" SpongeVanilla jar downloaded</p>
     * @param bungeeVersion <p>The version of the "latest" bungee jar downloaded</p>
     */
    public Server(String name, String path, boolean enabled, String typeName, String serverVersion, String maxRam,
            String vanillaVersion, String snapshotVersion, String spongeVanillaVersion, String bungeeVersion) {
        this.name = name;
        this.path = path;
        this.enabled = enabled;
        this.type = ServerTypeHandler.getByName(typeName);
        this.serverVersion = serverVersion;
        this.maxRam = maxRam;
        this.vanillaVersion = vanillaVersion;
        this.snapshotVersion = snapshotVersion;
        this.spongeVanillaVersion = spongeVanillaVersion;
        this.bungeeVersion = bungeeVersion;
        this.playerList = new ArrayList<>();
    }

    /**
     * Gets the name of the server
     *
     * @return <p>The name of the server</p>
     */
    public String getName() {
        return this.name;
    }

    /**
     * Whether the server has been started
     *
     * @return <p>True if the server has been started. False otherwise</p>
     */
    public boolean isStarted() {
        return started;
    }

    /**
     * Gets the name of the server type used by this server
     *
     * @return <p>The name of the server type used by this server</p>
     */
    public String getTypeName() {
        return this.type.getName();
    }

    /**
     * Gets the version used given server type used
     *
     * @return <p>The server version given server type</p>
     */
    public String getServerVersion() {
        return this.serverVersion;
    }

    public String getPath() {
        return this.path;
    }

    public Process getProcess() {
        return this.process;
    }

    public String getMaxRam() {
        return this.maxRam;
    }

    public static String[] getRamList() {
        return ramList;
    }

    public String getVanillaVersion() {
        return this.vanillaVersion;
    }

    public String getSnapshotVersion() {
        return this.snapshotVersion;
    }

    public String getSpongeVanillaVersion() {
        return this.spongeVanillaVersion;
    }

    public String getBungeeVersion() {
        return this.bungeeVersion;
    }

    public ArrayList<String> getPlayers() {
        return this.playerList;
    }

    public void stopped() {
        process = null;
        writer = null;
        reader = null;
        started = false;
    }

    /**
     * @return A representation of the name of a jarfile.
     */
    private String getType() {
        if (this.type.getName().equals("Custom")) {
            return this.serverVersion;
        } else {
            return "";
        }
    }

    public boolean isEnabled() {
        return this.enabled;
    }

    public void toggle(boolean value) {
        this.enabled = value;
    }

    public boolean hasPlayer(String name) {
        for (String player : this.playerList) {
            if (player.equals(name)) {
                return true;
            }
        }
        return false;
    }

    public void addPlayer(String name) {
        this.playerList.add(name);
        Profile.getGUI().addPlayer(name);
    }

    /**
     * Removes a player with the selected name from the playerlist.
     *
     * @param name The name of the player to remove
     */
    public void removePlayer(String name) {
        playerList.removeIf(player -> player.equals(name));
        Profile.getGUI().removePlayer(name);
    }

    public void setPath(String path) {
        this.path = path;
    }

    public void setType(ServerType type) {
        this.type = type;
    }

    public void setMaxRam(String ram) {
        this.maxRam = ram;
    }

    /**
     * Sets the server's server version to a valid version, or ignores the request.
     *
     * @param serverVersion Version number.
     */
    public void setServerVersion(String serverVersion) throws IllegalArgumentException {
        if (this.type.getName().equals("Custom")) {
            this.serverVersion = serverVersion;
        } else {
            String[] versions = this.type.getVersions();
            for (String version : versions) {
                if (version.equals(serverVersion)) {
                    this.serverVersion = serverVersion;
                    return;
                }
            }
            throw new IllegalArgumentException("Invalid server version.");
        }
    }

    /**
     * Tries to stop all enabled servers.
     *
     * @throws IOException If a writer's process is already closed but not null.
     */
    public static void stop() throws IOException {
        for (Collection collection : Profile.getCurrent().getCollections()) {
            Server server = collection.getServer();
            if (server.writer != null) {
                if (server.type.getName().equals("Bungee")) {
                    server.writer.write("end\n");
                } else {
                    server.writer.write("stop\n");
                }
                server.writer.flush();
                server.writer = null;
                server.started = false;
            }
        }
    }

    /**
     * Runs all enabled servers with their settings.
     */
    public static void startServers() {
        Profile.getGUI().setStatus("Starting servers");
        for (Collection collection : Profile.getCurrent().getCollections()) {
            if (!collection.getServer().run()) {
                Profile.getGUI().setStatus("An error occurred. Start aborted");
                try {
                    Server.stop();
                } catch (IOException e) {
                    e.printStackTrace();
                }
                Profile.getGUI().updateRunning(false);
                return;
            }
        }
    }

    /**
     * Runs a Minecraft server
     *
     * @return <p>True if nothing went wrong</p>
     */
    private boolean run() {
        if (!this.enabled) {
            this.started = false;
            return true;
        }
        this.started = true;
        if (!initializeJarDownload() || !delayStartup()) {
            return false;
        }

        try {
            startServerProcess();
            Profile.getGUI().setStatus("Servers are running");
            this.started = true;
            return true;
        } catch (IOException e) {
            Profile.getGUI().setStatus("Could not start server");
            this.started = false;
            return false;
        }
    }

    /**
     * Reads all available output from the server process
     *
     * @return <p>The server output</p>
     * @throws IOException <p>If reading from the reader fails</p>
     */
    public String readFromServer() throws IOException {
        String line;
        StringBuilder text = new StringBuilder();
        while (reader.ready() && (line = reader.readLine()) != null) {
            text.append(line).append("\n");
        }
        return text.toString().trim();
    }

    /**
     * Starts the process running this server
     *
     * @throws IOException <p>If the process cannot be started</p>
     */
    private void startServerProcess() throws IOException {
        ProcessBuilder builder;
        String serverPath;
        String serverFile;
        if (type.getName().equals("Custom")) {
            serverFile = serverVersion;
        } else {
            serverFile = this.type.getName() + serverVersion + ".jar";
        }
        if (Profile.getCurrent().getDownloadAllAvailableJARFiles() && !type.getName().equals("Custom")) {
            serverPath = jarDirectory + serverFile;
        } else {
            serverPath = this.path + File.separator + serverFile;
        }
        builder = new ProcessBuilder("java", "-Xmx" + this.maxRam, "-Xms512M",
                "-Djline.terminal=jline.UnsupportedTerminal", "-Dcom.mojang.eula.agree=true", "-jar", serverPath,
                "nogui");
        builder.directory(new File(this.path));
        builder.redirectErrorStream(true);
        this.process = builder.start();
        this.writer = new BufferedWriter(new OutputStreamWriter(this.process.getOutputStream()));
        this.reader = new BufferedReader(new InputStreamReader(this.process.getInputStream()));
    }

    /**
     * Delays the server's startup for the given amount of time
     *
     * @return <p>True if the delay was successful</p>
     */
    private boolean delayStartup() {
        try {
            Profile.getGUI().setStatus("Delaying startup");
            TimeUnit.SECONDS.sleep(Profile.getCurrent().getDelayStartup());
            return true;
        } catch (InterruptedException e) {
            e.printStackTrace();
            this.started = false;
            return false;
        }
    }

    /**
     * Starts downloading the necessary .jar file
     *
     * @return <p>True if nothing went wrong</p>
     */
    private boolean initializeJarDownload() {
        if (!Profile.getCurrent().getDownloadAllAvailableJARFiles()) {
            try {
                Profile.getGUI().setStatus("Downloading jar...");
                this.downloadJar();
                Profile.getGUI().setStatus("File downloaded");
            } catch (IOException e) {
                System.out.println(e.getMessage());
                Profile.getGUI().setStatus("Error: Jar file not found");
                e.printStackTrace();
                this.started = false;
                return false;
            }
        }
        return true;
    }

    /**
     * Downloads necessary .jar file for the server.
     *
     * @throws FileNotFoundException <p>If the file was not found and could not be acquired</p>
     */
    private void downloadJar() throws IOException {
        String path = this.path + File.separator;
        File file = new File(path + this.getType());
        if (!(file.isFile() || type.downloadJar(path, this.serverVersion))) {
            throw new FileNotFoundException("Jar file could not be downloaded.");
        }
    }

    /**
     * Sends a command to this server through its writer.
     *
     * @param command <p>Command to send to the server</p>
     * @throws IOException <p>If write fails</p>
     */
    public void sendCommand(String command) throws IOException {
        if (this.process != null && this.writer != null) {
            this.writer.write(command + "\n");
            this.writer.flush();
        }
    }
}