package net.knarcraft.minecraftserverlauncher.server.servertypes;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Paths;

import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.downloadFile;
import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.readFile;

public class Vanilla extends AbstractServerType {

    private final String versionURL;
    private String lastVersion;
    String releaseType;

    /**
     * Instantiates a vanilla server type
     *
     * @param typeName        <p>The name of this server type to display</p>
     * @param versions        <p>Available versions for this server type</p>
     * @param versionURL      <p>The URL used for downloading the version document</p>
     * @param downloadURL     <p>The URL used for downloading the new file</p>
     */
    public Vanilla(String typeName, String[] versions, String versionURL, String downloadURL) {
        super(typeName, versions, downloadURL);
        this.versionURL = versionURL;
        this.releaseType = "release";
    }

    @Override
    public boolean downloadJar(String folder, String version) throws IOException {
        String file = this.getName() + version + ".jar";
        File filePath = new File(folder + file);
        if (version.equals("Latest")) {
            String[] latestData = getLatestFile();
            String latest = latestData[0];
            String jarFile = latestData[1];
            String currentVersion = lastVersion;
            lastVersion = latest;
            return (filePath.isFile() && latest.equals(currentVersion)) || downloadFile(jarFile, Paths.get(filePath.toURI()));
        } else {
            String downloadURL = getVanillaDownloadURL(getServerFileVersionURL(version));
            return filePath.isFile() || downloadFile(downloadURL, Paths.get(filePath.toURI()));
        }
    }

    /**
     * Gets the URL to the .jar file for the newest version
     *
     * @return <p>An array containing the latest version and a link to its file</p>
     * @throws IOException <p>If the remote resource cannot be readFromServer</p>
     */
    private String[] getLatestFile() throws IOException {
        String versionText = readFile(versionURL);
        JsonObject jsonObject = new JsonParser().parse(versionText).getAsJsonObject();
        String latest = jsonObject.getAsJsonObject("latest").get(releaseType).getAsString();
        String versionURL = getServerFileVersionURL(latest);
        String jarURL = getVanillaDownloadURL(versionURL);
        return new String[]{latest, jarURL};
    }

    /**
     * Gets the URL necessary for downloading a given minecraft .jar file
     *
     * @param versionURL <p>The URL to the version document describing the .jar file</p>
     * @return <p>The URL necessary do download the .jar file</p>
     * @throws IOException <p>If the remote resource cannot be readFromServer</p>
     */
    private String getVanillaDownloadURL(String versionURL) throws IOException {
        String versionText = readFile(versionURL);
        JsonObject jsonObject = new JsonParser().parse(versionText).getAsJsonObject();
        return jsonObject.getAsJsonObject("downloads").getAsJsonObject("server").get("url").getAsString();
    }

    /**
     * Gets the version URL from the Minecraft vanilla version document
     *
     * @param targetVersion <p>The version to download</p>
     * @return <p>The URL to the file</p>
     * @throws IOException <p>If the file cannot be downloaded</p>
     */
    private String getServerFileVersionURL(String targetVersion) throws IOException {
        String versionText = readFile(versionURL);
        JsonObject jsonObject = new JsonParser().parse(versionText).getAsJsonObject();
        JsonArray availableVersions = jsonObject.getAsJsonArray("versions");
        for (JsonElement availableVersion : availableVersions) {
            JsonObject versionObject = availableVersion.getAsJsonObject();
            if (!versionObject.get("id").getAsString().equals(targetVersion)) {
                continue;
            }
            return versionObject.get("url").getAsString();
        }
        throw new FileNotFoundException("Unable to find the requested Minecraft vanilla .jar file.");
    }

}
